'''
2019 Bjoern Annighoefer
'''

from ..value import VAL, NON, LST
from ..error import EOQ_ERROR_INVALID_VALUE

DATE_TIME_STR_FORMAT = "%m/%d/%Y %H:%M:%S"

'''
    A class indicating that nothing more to process here
'''
class TRM(VAL):
    def __init__(self,v):
        self.v = v
    def __eq__(self, other):
        if(isinstance(other, TRM)):
            return self.v == other.v
        else:
            return self.v == other

def Determinate(res:VAL) -> VAL:
    if(IsList(res)):
        res = LST([Determinate(r) for r in res])
    elif(isinstance(res, TRM)):
        res = res.v
    return res

'''
    Operations on multiple objects, structures and lists
    
'''

def ApplyToAllElements(context,functor):
    res = NON()
    if(IsList(context)):
        res = LST([ApplyToAllElements(c,functor) for c in context])
    elif(isinstance(context, TRM)):
        res = res
    else:
        res = functor(context)
    return res


def ApplyToAllElementsInA(a:VAL, b:VAL,functor) -> VAL:
    res = NON()
    if(IsList(a)):
        res = LST([ApplyToAllElementsInA(c,b,functor) for c in a])
    elif(isinstance(a, TRM)):
        res = a
    else:
        res = functor(a,b)
    return res

def ApplyToAllElementsInB(a:VAL, b:VAL, functor) -> VAL:
    res = NON()
    if(IsList(b)):
        res = LST([ApplyToAllElementsInB(a,c,functor) for c in b])
    else:
        res = functor(a,b)
    return res

def ApplyToAllListsOfElementsInA(a,b,functor):
    res = NON()
    if(IsListOfObjects(a)):
        res = functor(a,b)
    else:
        res = LST([ApplyToAllListsOfElementsInA(c,b,functor) for c in a])
    return res

def ApplyToAllListsOfElementsInB(a,b,functor):
    res = NON()
    if(IsListOfObjects(b)):
        res = functor(a,b)
    else:
        res = LST([ApplyToAllListsOfElementsInB(a,c,functor) for c in b])
    return res

def ApplyToSimilarListsOfObjects(op1,op2,listVsListFunc,listVsStructFunc,structVsListOp,param=None):
    res = NON()
    if(IsListOfObjects(op1) and IsListOfObjects(op2)):
        res = listVsListFunc(op1,op2,param)
    elif(IsListOfObjects(op1)):
        res = listVsStructFunc(op1,op2,param)
    elif(IsListOfObjects(op2)):
        res = structVsListOp(op1,op2,param)
    elif(len(op1) == len(op2)):
        res = LST([ApplyToSimilarListsOfObjects(op1[i],op2[i],listVsListFunc,listVsStructFunc,structVsListOp,param) for i in range(len(op1))])
    else:
        raise EOQ_ERROR_INVALID_VALUE("Non comparable element list structures detected.")
    return res

def ApplyToSimilarElementStrutures(op1,op2,elemVsElemFunc,elemVsStruct,structVsElemOp,param=None):
    res = NON()
    if(IsNoList(op1) and IsNoList(op2)):
        res = elemVsElemFunc(op1,op2,param)
    elif(IsNoList(op1)):
        res = elemVsStruct(op1,op2,param)
    elif(IsNoList(op2)):
        res = structVsElemOp(op1,op2,param)
    elif(len(op1) == len(op2)):
        res = LST([ApplyToSimilarElementStrutures(op1[i],op2[i],elemVsElemFunc,elemVsStruct,structVsElemOp,param) for i in range(len(op1))])
    else:
        raise EOQ_ERROR_INVALID_VALUE("Non comparable element structures detected.")
    return res

def IsList(val:VAL) -> bool:
    return LST == type(val)

def IsNoList(val:VAL) -> bool:
    return LST != type(val)

def IsListOfObjects(obj:VAL) -> bool:
    """Check if obj is a list of elements, but not a list of lists.
    """
    if(IsList(obj)):
        if(len(obj)==0): 
            return True
        for o in obj:
            if IsList(o):
                return False #any non list object make the search fail
        return True
    return False

def ShowProgress(progress):
    print('Total progress: %d%%'%(progress))
    return



'''
2024 Bjoern Annighoefer
'''

from ..domain import Domain
from ..command import Cmp
from ..query import His, Obj
from ..value import BOL,U32,I32,I64,F32,LST,NON,STR

def CreateWarehouseM2Model(domain:Domain, sessionId:str=None)->Obj:
    """Create warehouse meta-model as given in the getting started section
    
                                                              _______________
         ___________              _______________            |    Product    |
        | Warehouse | categories |   Category    |  products |---------------|
        |-----------|<>--------->|---------------|<>-------->| name : string |
        |           |           *| name : string |          *| price : real  |
        |___________|        o-->|_______________|<>-o       | sells : int   |
                             |  *                    |       |_______________|
                             o-----------------------o
                                   categories
        
    """
    #the following is generated form eoq2's example warehouse.ecore
    cmd = Cmp().Crt(STR('*M2MODEL'),U32(1),LST([STR('warehouse')]),NON(),LST([]),resName='o0')\
               .Crt(STR('*M2CLASS'),U32(1),LST([STR('Warehouse'),BOL(False),His(STR('o0'))]),NON(),LST([]),resName='o1')\
               .Crt(STR('*M2CLASS'),U32(1),LST([STR('Category'),BOL(False),His(STR('o0'))]),NON(),LST([]),resName='o2')\
               .Crt(STR('*M2CLASS'),U32(1),LST([STR('Product'),BOL(False),His(STR('o0'))]),NON(),LST([]),resName='o3')\
               .Crt(STR('*M2COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o1')),His(STR('o2')),I64(-1),BOL(False)]),NON(),LST([]),resName='o4')\
               .Crt(STR('*M2ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o2')),STR('*STR'),I64(1),NON(),NON()]),NON(),LST([]),resName='o5')\
               .Crt(STR('*M2COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o2')),His(STR('o2')),I64(-1),BOL(False)]),NON(),LST([]),resName='o6')\
               .Crt(STR('*M2COMPOSITION'),U32(1),LST([STR('products'),His(STR('o2')),His(STR('o3')),I64(-1),BOL(False)]),NON(),LST([]),resName='o7')\
               .Crt(STR('*M2ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o3')),STR('*STR'),I64(1),NON(),NON()]),NON(),LST([]),resName='o8')\
               .Crt(STR('*M2ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o3')),STR('*F32'),I64(1),NON(),NON()]),NON(),LST([]),resName='o9')\
               .Crt(STR('*M2ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o3')),STR('*I32'),I64(1),NON(),NON()]),NON(),LST([]),resName='o10')
    #execute the command on the domain
    res = domain.Do(cmd,sessionId)
    m1model = res[0]
    return m1model
          
def CreateWarehouseM1Model(domain:Domain, sessionId=None)->Obj:
    """Create warehouse user model with the following content: 
    
        <warehouse>
          <categories name="food">
            <categories name="fresh">
              <categories name="fruit">
                <products name="Apple" sells="400" price="0.25"/>
                <products name="Banana" sells="1" price="0.35"/>
                <products name="Orange" sells="20" price="0.45"/>
                <products name="Melon" sells="2" price="1.1"/>
              </categories>
              <categories name="vegetables">
                <products name="Cucumber" sells="4" price="0.5"/>
                <products name="Lettuce" sells="2" price="0.8"/>
              </categories>
              <categories name="milk">
                <products name="Cream" sells="300" price="0.5"/>
                <products name="Milk" sells="700" price="0.6"/>
                <categories name="cheese">
                  <products name="Goat Cheese" price="5.0"/>
                  <products name="Gouda" sells="100" price="2.0"/>
                </categories>
              </categories>
            </categories>
            <categories name="drinks">
              <products name="Coffee" sells="100" price="2.0"/>
              <products name="Hot Chocolate" sells="1" price="2.5"/>
              <categories name="lemonade">
                <products name="Coke" sells="1000" price="1.0"/>
                <products name="Energy Drink" price="10.0"/>
              </categories>
            </categories>
          </categories>
        </warehouse>
    
    """
    #the following is generated form eoq2's example GeneralsStore.warehouse
    cmd = Cmp().Crt(STR('*M1MODEL'),U32(1),LST([STR('warehouse'),STR('m1model')]),NON(),LST([]),resName='o0')\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Warehouse'),His(STR('o0')),NON()]),NON(),LST([]),resName='o1')\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('food')]),NON(),LST([]),resName='o2')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o2')),STR('food')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('fresh')]),NON(),LST([]),resName='o3')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o3')),STR('fresh')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('fruit')]),NON(),LST([]),resName='o4')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o4')),STR('fruit')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Apple')]),NON(),LST([]),resName='o5')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o5')),STR('Apple')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o5')),F32(0.250000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o5')),I32(400)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o4')),His(STR('o5'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Banana')]),NON(),LST([]),resName='o6')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o6')),STR('Banana')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o6')),F32(0.350000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o6')),I32(1)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o4')),His(STR('o6'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Orange')]),NON(),LST([]),resName='o7')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o7')),STR('Orange')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o7')),F32(0.450000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o7')),I32(20)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o4')),His(STR('o7'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Melon')]),NON(),LST([]),resName='o8')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o8')),STR('Melon')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o8')),F32(1.100000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o8')),I32(2)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o4')),His(STR('o8'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o3')),His(STR('o4'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('vegetables')]),NON(),LST([]),resName='o9')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o9')),STR('vegetables')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Cucumber')]),NON(),LST([]),resName='o10')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o10')),STR('Cucumber')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o10')),F32(0.500000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o10')),I32(4)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o9')),His(STR('o10'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Lettuce')]),NON(),LST([]),resName='o11')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o11')),STR('Lettuce')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o11')),F32(0.800000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o11')),I32(2)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o9')),His(STR('o11'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o3')),His(STR('o9'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('milk')]),NON(),LST([]),resName='o12')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o12')),STR('milk')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('cheese')]),NON(),LST([]),resName='o13')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o13')),STR('cheese')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Goat Cheese')]),NON(),LST([]),resName='o14')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o14')),STR('Goat Cheese')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o14')),F32(5.000000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o14')),I32(0)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o13')),His(STR('o14'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Gouda')]),NON(),LST([]),resName='o15')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o15')),STR('Gouda')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o15')),F32(2.000000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o15')),I32(100)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o13')),His(STR('o15'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o12')),His(STR('o13'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Cream')]),NON(),LST([]),resName='o16')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o16')),STR('Cream')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o16')),F32(0.500000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o16')),I32(300)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o12')),His(STR('o16'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Milk')]),NON(),LST([]),resName='o17')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o17')),STR('Milk')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o17')),F32(0.600000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o17')),I32(700)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o12')),His(STR('o17'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o3')),His(STR('o12'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o2')),His(STR('o3'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('drinks')]),NON(),LST([]),resName='o18')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o18')),STR('drinks')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Category'),His(STR('o0')),STR('lemonade')]),NON(),LST([]),resName='o19')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o19')),STR('lemonade')]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Coke')]),NON(),LST([]),resName='o20')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o20')),STR('Coke')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o20')),F32(1.000000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o20')),I32(1000)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o19')),His(STR('o20'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Energy Drink')]),NON(),LST([]),resName='o21')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o21')),STR('Energy Drink')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o21')),F32(10.000000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o21')),I32(0)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o19')),His(STR('o21'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o18')),His(STR('o19'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Coffee')]),NON(),LST([]),resName='o22')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o22')),STR('Coffee')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o22')),F32(2.000000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o22')),I32(100)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o18')),His(STR('o22'))]),NON(),LST([]))\
               .Crt(STR('*M1OBJECT'),U32(1),LST([STR('warehouse__Product'),His(STR('o0')),STR('Hot Chocolate')]),NON(),LST([]),resName='o23')\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('name'),His(STR('o23')),STR('Hot Chocolate')]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('price'),His(STR('o23')),F32(2.500000)]),NON(),LST([]))\
               .Crt(STR('*M1ATTRIBUTE'),U32(1),LST([STR('sells'),His(STR('o23')),I32(1)]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('products'),His(STR('o18')),His(STR('o23'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o2')),His(STR('o18'))]),NON(),LST([]))\
               .Crt(STR('*M1COMPOSITION'),U32(1),LST([STR('categories'),His(STR('o1')),His(STR('o2'))]),NON(),LST([]))
    #execute the command on the domain
    res = domain.Do(cmd,sessionId)
    m1model = res[0]
    return m1model


'''
2019 Bjoern Annighoefer
'''

from .logger import Logger, RegisterLogger, DEFAULT_LOGGER_LEVELS

class NoLogger(Logger):
    '''A logger which does nothing
    '''
    def __init__(self, activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        super().__init__(activeLevels)
        
    #@Override
    def ShallLog(self):
        return False

    #@Override
    def _Log(self, levelName:str, msg:str):
        pass #do nothing

RegisterLogger("NOL",NoLogger,[])
    
    


from .command import\
    RegisterCmdType,\
    CmdFactory,\
    CMD_SEP,\
    CMD_ARG_SEP,\
    CMD_TYPES,\
    DEL_MODES,\
    CLO_MODES,\
    MRG_MODES,\
    EVT_TYPES,\
    EVT_KEYS,\
    Cmd,\
    Get,\
    Upd,\
    Del,\
    Set,\
    Add,\
    Rem,\
    Mov,\
    Clo,\
    Crt,\
    Cpr,\
    Mrg,\
    Hel,\
    Ses,\
    Gby,\
    Sts,\
    Chg,\
    Obs,\
    Ubs,\
    Msg,\
    Ver,\
    Gac,\
    Res,\
    Err,\
    Evt,\
    UpdEvt,\
    CrtEvt,\
    DelEvt,\
    MsgEvt,\
    WatEvt,\
    VecEvt,\
    Scc,\
    Cus,\
    Cmp,\
    IsReadOnlyCmd,\
    PrmCmd,\
    CmdInfo,\
    GetCmdInfo



'''
2019 Bjoern Annighoefer
'''

from .logger import Logger,RegisterLogger, DEFAULT_LOGGER_LEVELS, LOG_LEVELS_NAME_LUT

import os
import logging

class FileLogger(Logger):
    """A logger that logs to a file.
    If split files is desired, each log level has its own file.
    The log files are created in the logDir with the prefix and the log level as name.
    """
    def __init__(self,activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING,logDir:str='./log',prefix:str='log',splitFiles:bool=False):
        super().__init__(activeLevels)
        self.logDir = logDir
        self.prefix = prefix #the text added as log file name
        self.splitFiles = splitFiles
        #internals
        self.pyLoggers = {}
        #make sure the dir exists
        if(not os.path.isdir(self.logDir)):
            os.makedirs(self.logDir)
        #create loggers
        if(splitFiles):
            # create native python loggers for each level
            for k,v in LOG_LEVELS_NAME_LUT.items():
                if(k & activeLevels):
                    self._InitPyLogger(v,v)
        else:
            # create native python logger
            self._InitPyLogger("all","")

    def _InitPyLogger(self,name,postfix):
        """Initializes a python logger for the given name
        and adds it to the pyLoggers dict.
        """
        logger = logging.getLogger(name)
        logFile = os.path.join(self.logDir, "%s%s.log" % (self.prefix, postfix))
        fh = logging.FileHandler(logFile, 'w')
        fh.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(message)s')
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        logger.setLevel(logging.INFO)
        self.pyLoggers[name] = logger

    #@Override         
    def _Log(self,levelName:str,msg:str):
        if(self.splitFiles):
            self.pyLoggers[levelName].info(msg)
        else:
            self.pyLoggers["all"].info(f"[{levelName}] {msg}")


RegisterLogger("FIL",FileLogger,[ "logDir","prefix","splitFiles"])



'''
 Bjoern Annighoefer 2024
'''

from ...domain import Domain
from .cmdstream import EoqFileInStream, ToDomainCmdStream


def LoadEoqFile(infile: str, domain: Domain = None, sessionId: str = None, validateBeforeLoad: bool = True) -> None:
    """Loads an EOQ file to a specified domain.

    Args:
        infile (str): Path to the EOQ file to be loaded.
        domain (Domain, optional): The domain to load the EOQ file to. Defaults to None.
        sessionId (str, optional): Session ID for the domain. Defaults to None.
        validateBeforeLoad (bool, optional): Whether to validate the EOQ file before loading. Defaults to True.

    Raises:
        [Exception]: If an error occurs during loading or validation.

    This function loads an EOQ file to a specified domain. If no domain is provided, a default interpreter will be used.
    The EOQ file can be optionally validated before loading.
    """
    fileInStream = EoqFileInStream()
    if(validateBeforeLoad):
        #load the file, before the domain is connected.
        fileInStream.Begin()
        fileInStream.LoadEoqFile(infile)
        fileInStream.Flush()
    #connect the domain and load the file (again)
    fileInStream.Connect(ToDomainCmdStream(domain,sessionId,False,True))
    fileInStream.Begin()
    fileInStream.LoadEoqFile(infile)
    fileInStream.Flush()


"""
 2022 Bjoern Annighoefer
"""
from .remotedomainbasea import RemoteDomainBaseA
from .serialdomain import SerialDomain

from ..domain import Domain
from ..frame import Frm, FRM_TYPE
from ..command import Cmp, Err
from ..config import Config
from ..error import EOQ_ERROR, EOQ_ERROR_CODES
from ..error import EOQ_ERROR_INVALID_TYPE

class RemoteDomainServer(RemoteDomainBaseA):
    def __init__(self, domain:Domain, shallForwardSerializedCmds:bool, config:Config):
        super().__init__(config)
        self.domain = domain
        if(shallForwardSerializedCmds and not isinstance(domain,SerialDomain)):
            raise EOQ_ERROR_INVALID_TYPE('Can only forward serial commands to SerialDomains.')
        self.shallForwardSerializedCmds = shallForwardSerializedCmds
        # self.config = config
        # self.remoteFrmTxSerializer = CreateSerializer(config.remoteFrmTxSerializer)
        # self.remoteCmdTxSerializer = CreateSerializer(config.remoteCmdTxSerializer)
        self.CmdFrameHandler = self.__ProcesscmdFrameNormal
        #prepare processor of commands
        if(shallForwardSerializedCmds):
            self.CmdFrameHandler = self.__ProcesscmdFrameSerial
    
    def OnSerFrmReceived(self, frmStr:str):
        frm = self.remoteFrmRxSerializer.DesFrm(frmStr)
        self.OnFrmReceived(frm)
    
    def OnFrmReceived(self, frm:Frm):
        sessionId = frm.sid
        if(FRM_TYPE.CMD == frm.typ):
            resFrm = self.CmdFrameHandler(frm)
            resFrmStr = self.remoteFrmTxSerializer.SerFrm(resFrm)
            self.SendSerFrm(resFrmStr,sessionId)
        elif(FRM_TYPE.OBS == frm.typ):
            self.domain.Observe(self.__OnDomainEvent, sessionId, sessionId)
        elif(FRM_TYPE.UBS == frm.typ):
            self.domain.Unobserve(self.__OnDomainEvent, sessionId, sessionId)
        else:
            self.logger.Warn("Skipped unexpected frm type: %s"%(frm.typ))

    def __OnDomainEvent(self, evts, context, source)->None:
        #print('Forwarding %d events from the domain for context %s and source %s.'%(len(evts),context,source))
        evtCmd = Cmp(evts) #pack commands in compound
        evtCmdStr = self.remoteCmdTxSerializer.SerCmd(evtCmd)
        frm = Frm(FRM_TYPE.EVT, 0, self.remoteCmdTxSerializer.Name(), evtCmdStr, context) #context equals the session id of the receiver. Source can is the session id of the event source4
        frmStr = self.remoteFrmTxSerializer.SerFrm(frm)
        self.SendSerFrm(frmStr,context)
        
    ### INTERFACE ###
    
    def SendSerFrm(self, frmStr:str, sessionId:str)->None:
        """Needs to be overwritten by the implementation
        """
        raise NotImplemented()

    def __ProcesscmdFrameNormal(self, frm:Frm)->Frm:
        """returns the result frame, by invoking a normal domain
        """
        commandId = frm.uid
        serType = frm.ser
        try:
            cmd = self._DesCmd(serType,frm.dat)
            res = self.domain.RawDo(cmd,frm.sid,frm.roc)
        except EOQ_ERROR as e:
            res = Err(e.code,e.msg)
        except Exception as e:
            res = Err(EOQ_ERROR_CODES.UNKNOWN,str(e))
        resStr = self._SerCmd(serType,res) #reply with the same serialization
        resFrm = Frm(FRM_TYPE.RES, commandId, serType, resStr, frm.sid)
        return resFrm
    
    def __ProcesscmdFrameSerial(self, frm:Frm)->Frm:
        """returns the result frame by invoking a serial domain
        """
        resStr = None
        try:
            (resStr,resSer) = self.domain.SerRawDo(frm.dat,frm.sid,frm.ser,frm.roc)
        except EOQ_ERROR as e:
            res = Err(e.code,e.msg)
            resStr = self.remoteCmdTxSerializer.SerCmd(res)
            resSer = self.remoteCmdTxSerializer.Name()
        except Exception as e:
            res = Err(EOQ_ERROR_CODES.UNKNOWN,str(e))
            resStr = self.remoteCmdTxSerializer.SerCmd(res)
            resSer = self.remoteCmdTxSerializer.Name()
        resFrm = Frm(FRM_TYPE.RES, frm.uid, resSer, resStr, frm.sid)
        return resFrm

    ### DOMAIN WRAPPING ###

    # @Override
    def RawDo(self, cmd, sessionId:str, readOnly:bool)->object:
        """Allow to access the inner domain
        """
        return self.domain.RawDo(cmd, sessionId, readOnly)